#!/usr/bin/perl

package bvlshr_const_by_bvmul;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(benchmarks_bvlshr_const_by_bvmul);

use POSIX qw/floor/;
use List::Util qw[max];

use base_ops;

###################################################
# Right shift
###################################################

sub bvmul {
	@_ == 2 || die("ERROR: too few/many arguments for bvmul()\n\tusage: bvmul(term1,term2)\n");

	local $term1 = $_[0];
	local $term2 = $_[1];

	return "(bvmul $term1 $term2)";
}

# double-powers of 2, i.e. 2^2^n
local $dpows2 = ();

sub generate_dpows2 {
	@_ == 2 || die("ERROR: too few/many arguments for generate_dpows2()\n\tusage: generate_dpows2(n, bw)\n");

	local $n = $_[0];
	local $bw = $_[1];
	
	$dpows2[0] = declareVar("dpow2_0", $bw);
	assert(
		equal(
			$dpows2[0],
			bvconst(2, $bw)
		));
	
	local $i;
	for ($i = 1; $i <= $n; $i++) {
		$dpows2[$i] = declareVar("dpow2_$i", $bw);
		assert(
			equal(
				$dpows2[$i],
				bvmul($dpows2[$i-1], $dpows2[$i-1])
		));
	}
}

# power of 2, i.e. 2^n
sub pow2 {
	@_ == 2 || die("ERROR: too few/many arguments for pow2()\n\tusage: pow2(n, bw)\n");

	local $n = $_[0];
	local $bw = $_[1];
	
	local $result = (($n & 1) == 0) ? bvconst(1, $bw) : $dpows2[0];
	
	$n >>= 1;
	
	local $i;
	for ($i = 1; $n > 0; $i++) {
		if (($n & 1) > 0) {
			$result = bvmul($result, $dpows2[$i]);
		}
		
		$n >>= 1;
	}
	
	return $result;
}

# double-powers of 2 minus 1, i.e. 2^2^n-1
local $dpows2_minus_1 = ();

sub generate_dpows2_minus_1 {
	@_ == 2 || die("ERROR: too few/many arguments for generate_dpows2_minus_1()\n\tusage: generate_dpows2_minus_1(n, bw)\n");

	local $n = $_[0];
	local $bw = $_[1];
	
	$dpows2_minus_1[0] = declareVar("dpow2_0_minus_1", $bw);
	assert(
		equal(
			$dpows2_minus_1[0],
			bvconst(1, $bw)
		));
	
	local $i;
	for ($i = 1; $i <= $n; $i++) {
		$dpows2_minus_1[$i] = declareVar("dpow2_".$i."_minus_1", $bw);
		assert(
			equal(
				$dpows2_minus_1[$i],
				bvor(
					$dpows2_minus_1[$i-1],
					bvmul(
						$dpows2_minus_1[$i-1],
						$dpows2[$i-1])
		)));
	}
}

# power of 2 minus 1, i.e. 2^n-1
sub pow2_minus_1 {
	@_ == 2 || die("ERROR: too few/many arguments for pow2_minus_1()\n\tusage: pow2_minus_1(n, bw)\n");

	local $n = $_[0];
	local $bw = $_[1];
	
	local $result = (($n & 1) == 0) ? bvconst(0, $bw) : $dpows2_minus_1[0];
	
	$n >>= 1;
	
	local $i;
	for ($i = 1; $n > 0; $i++) {
		if (($n & 1) > 0) {
			$result = bvor(
				bvmul(
					$result,
					$dpows2[$i]
				),
				$dpows2_minus_1[$i]
			);
		}
		
		$n >>= 1;
	}
	
	return $result;
}

###################################################
# Right shift by constant

sub bvlshr_const_by_bvmul {
	@_ == 3 || die("ERROR: too few/many arguments for bvlshr_const_by_bvmul()\n\tusage: bvlshr_const_by_bvmul(term,const,bw)\n");
  
	local $term = $_[0];
	local $const = $_[1];
	local $bw = $_[2];

# 	if ($const == 0) {
# 		return $term;
# 	}
	if ($const < $bw) {
		local $exp = max($const, $bw-$const);
		generate_dpows2(floor(log2($exp)), $bw);
		generate_dpows2_minus_1(floor(log2($exp)), $bw);
		
		local $d = declareVar("d", $bw);
		
		assert(
			equal(
				bvor(
					bvnot($d),
					pow2_minus_1($bw-$const, $bw)
				),
				bvnot(bvconst(0, $bw))
		));
		
		assert(
			equal(
				bvor(
					bvnot($term),
					pow2_minus_1($const, $bw)
				),
				bvnot(
					bvmul(
						$d,
						pow2($const, $bw)
		))));
	
		return $d;
	}
	else {
		return bvconst(0, $bw);
	}
}


sub test_bvlshr_const_by_bvmul {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvlshr_const_by_bvmul()\n\tusage: test_bvlshr_const_by_bvmul(input1,input2,bw)\n");


	# START

	startSmt("bvlshr", "bvmul");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	$input1 = declareVar($input1, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvlshr $input1 (_ bv$input2 $bw))"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvlshr_const_by_bvmul($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvlshr_const_by_bvmul {
	@_ == 6 || die("ERROR: too few/many arguments for benchmarks_bvlshr_const_by_bvmul()\n\tusage: benchmarks_bvlshr_const_by_bvmul(dir, bw_min, bw_max, bw_step, const_factor_min, const_factor_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];
	local $const_factor_min = $_[4];
	local $const_factor_step = $_[5];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw, $const_factor, $const;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(5))+1; $bw<=2000; $bw+=int(rand(200))) {
		for ($const_factor = $const_factor_min; $const_factor < 1; $const_factor += $const_factor_step) {
			$const = int($const_factor * $bw);

			println("Generating benchmark for bvlshr_const_by_bvmul with const $const and bit-width $bw...");
		
			$filename = "$dir/bvlshr_const_by_bvmul_".sprintf("%04d", $bw)."_".sprintf("%04d", $const).".smt2";
			open FILE, ">", $filename or die $!;
		
			select FILE;
		
			test_bvlshr_const_by_bvmul("input1", $const, $bw);
		
			select STDOUT;
		
			close FILE;
		}
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 5 ) {
		print "usage: bvlshr_const_by_bvmul  dir  bw_min  bw_max  bw_step  const_factor_min  const_factor_step\n";
		exit;
	}
	benchmarks_bvlshr_const_by_bvmul($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3], $ARGV[4], $ARGV[5]);
}
