#!/usr/bin/perl

package bvshl_const_by_bvmul;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(benchmarks_bvshl_const_by_bvmul);

use POSIX qw/floor/;

use base_ops;

###################################################
# Left shift
###################################################

sub bvmul {
	@_ == 2 || die("ERROR: too few/many arguments for bvmul()\n\tusage: bvmul(term1,term2)\n");

	local $term1 = $_[0];
	local $term2 = $_[1];

	return "(bvmul $term1 $term2)";
}

# double-powers of 2, i.e. 2^2^n
local $dpows2 = ();

sub generate_dpows2 {
	@_ == 2 || die("ERROR: too few/many arguments for generate_dpows2()\n\tusage: generate_dpows2(n, bw)\n");

	local $n = $_[0];
	local $bw = $_[1];
	
	$dpows2[0] = declareVar("dpow2_0", $bw);
	assert(
		equal(
			$dpows2[0],
			bvconst(2, $bw)
		));
	
	local $i;
	for ($i = 1; $i <= $n; $i++) {
		$dpows2[$i] = declareVar("dpow2_$i", $bw);
		assert(
			equal(
				$dpows2[$i],
				bvmul($dpows2[$i-1], $dpows2[$i-1])
		));
	}
}

# power of 2, i.e. 2^n
sub pow2 {
	@_ == 2 || die("ERROR: too few/many arguments for pow2()\n\tusage: pow2(n, bw)\n");

	local $n = $_[0];
	local $bw = $_[1];
	
	local $result = (($n & 1) == 0) ? bvconst(1, $bw) : $dpows2[0];
	
	$n >>= 1;
	
	local $i;
	for ($i = 1; $n > 0; $i++) {
		if (($n & 1) > 0) {
			$result = bvmul($result, $dpows2[$i]);
		}
		
		$n >>= 1;
	}
	
	return $result;
}

###################################################
# Left shift by constant

sub bvshl_const_by_bvmul {
	@_ == 3 || die("ERROR: too few/many arguments for bvshl_const_by_bvmul()\n\tusage: bvshl_const_by_bvmul(term,const,bw)\n");
  
	local $term = $_[0];
	local $const = $_[1];
	local $bw = $_[2];

	if ($const < $bw) {
		if ($const > 0) {
			generate_dpows2(floor(log2($const)), $bw);
		}
	
		return bvmul(
			$term,
			pow2($const, $bw)
		);
	}
	else {
		return bvconst(0, $bw);
	}
}


sub test_bvshl_const_by_bvmul {
	@_ == 3 || die("ERROR: too few/many arguments for test_bvshl_const_by_bvmul()\n\tusage: test_bvshl_const_by_bvmul(input1,input2,bw)\n");


	# START

	startSmt("bvshl by const", "bvmul");


	### INPUT VARIABLES

	local $input1 = $_[0];
	local $input2 = $_[1];
	local $bw = $_[2];

	$input1 = declareVar($input1, $bw);


	### RESULT BY SMT SOLVER

	local $result = declareVar("result", $bw);
	assert(
		equal(
			$result,
			"(bvshl $input1 (_ bv$input2 $bw))"
	));


	### COMPARE RESULT BY OUR APPROACH

	assert(
		lnot(
			equal(
				$result,
				bvshl_const_by_bvmul($input1, $input2, $bw)
	)));


	# FINISH

	finishSmt();
}


sub benchmarks_bvshl_const_by_bvmul {
	@_ == 6 || die("ERROR: too few/many arguments for benchmarks_bvshl_const_by_bvmul()\n\tusage: benchmarks_bvshl_const_by_bvmul(dir, bw_min, bw_max, bw_step, const_factor_min, const_factor_step)\n");
	local $dir = $_[0];
	local $bw_min = $_[1];
	local $bw_max = $_[2];
	local $bw_step = $_[3];
	local $const_factor_min = $_[4];
	local $const_factor_step = $_[5];

	unless (-d $dir) {
		mkdir $dir || die $!;
	}

	local $filename;

	local $bw, $const_factor, $const;
	for ($bw = $bw_min; $bw <= $bw_max; $bw += $bw_step) {
#	for ($bw = int(rand(50))+1; $bw<=800; $bw+=int(rand(100))) {
		for ($const_factor = $const_factor_min; $const_factor < 1; $const_factor += $const_factor_step) {
			$const = int($const_factor * $bw);
			
			println("Generating benchmark for bvshl_const_by_bvmul with const $const and bit-width $bw...");
		
			$filename = "$dir/bvshl_const_by_bvmul_".sprintf("%04d", $bw)."_".sprintf("%04d", $const).".smt2";
			open FILE, ">", $filename or die $!;
		
			select FILE;
		
			test_bvshl_const_by_bvmul("input1", $const, $bw);
		
			select STDOUT;
		
			close FILE;
		}
	}
}


#======= MAIN FUNCTION

__PACKAGE__->main(@ARGV) unless caller;

sub main {
	if ($#ARGV != 5 ) {
		print "usage: bvshl_const_by_bvmul  dir  bw_min  bw_max  bw_step  const_factor_min  const_factor_step\n";
		exit;
	}
	benchmarks_bvshl_const_by_bvmul($ARGV[0], $ARGV[1], $ARGV[2], $ARGV[3], $ARGV[4], $ARGV[5]);
}
